// MAT259 Assignment
////////////////////////////////////////////////////////////////////////////////
// 3D Visualization: Hans Zimmer's OST and Its Original Film Checkout Trend
//
// Author: Sihwa Park (sihwapark@mat.ucsb.edu)
// 1st Version : 02/22/2018
////////////////////////////////////////////////////////////////////////////////


// Future work:
// - animation accorind to time or interaction
// - show checkout count and date of each box/cylinder according to the mouse-over or click interaction
// - adding dvd/ost cover images (click event or only for the align mode)
// - add the align mode that aligns all checkout trend timelines in the same plane

import peasy.*;
import peasy.org.apache.commons.math.geometry.*;
import controlP5.*;

ControlP5 cp5;
PeasyCam cam;
Table table;
RadioButton selectItemRadioButton;
RadioButton selectGraphTypeRadioButton;

int rows, columns;
float[][] data;
float[] maxValues, minValues;
String[] titles;
String[] startDate;
int[] startingMonth;
int[] startingYear;
float[] rotationAngle;

PVector[] startPosition;
int[] indicesForStartingMonth;

int startYear;
int months;
int itemCount;
float boxMaxHeight = 50;
float boxWidth = 10;
float constantHorizontalMargin = 110;
float constantVerticalMargin = 100;

ArrayList<PVector> spiralTimeLine;

ArrayList<PVector> halfCylinderMeshTop = new ArrayList<PVector>();
ArrayList<PVector> halfCylinderMeshBottom = new ArrayList<PVector>();
ArrayList<PVector> halfCylinderMeshBody = new ArrayList<PVector>();
CameraState[] camStates;


final int BAR_GRAPH = 0;
final int CIRCLE_GRAPH = 1;

int mode = BAR_GRAPH;
//color[] itemColors = {#e84a5f, #45ada8};
color[] itemColors = {#f67280, #355c7d};
color backgroundColor = #2a363b;
String title = "Hans Zimmer's OST and Its Original Film Checkout Trend";
String legendDescription = "Ratio with each medium's\na max checkout";

int selectedIndex = 0;
boolean shapeTextMode = true;

void setup() {
  size(1700, 1050, P3D);
  smooth();
  
  cam = new PeasyCam(this, 1000);
  cam.lookAt(0, 0, 0);
  
  surface.setResizable(true);
  
  table = loadTable("HansOSTandMovie.csv");
  rows = table.getRowCount();
  columns = table.getColumnCount();
  println("Rows: " + rows + " Columns: " + columns);

  startYear =  table.getInt(1, 0);
  
  data = new float[columns - 2][rows - 1];
  itemCount = (columns - 2) / 2;
  maxValues = new float[itemCount * 2];
  minValues = new float[itemCount * 2];
  titles = new String[itemCount * 2];
  startingMonth = new int[itemCount];
  startingYear = new int[itemCount];
  startDate = new String[itemCount];
  startPosition = new PVector[itemCount];
  indicesForStartingMonth = new int[itemCount];
  months = (rows - 1);
  rotationAngle = new float[itemCount];
  
  
  camStates = new CameraState[itemCount + 1];
  
  for(int i = 0; i < columns - 2; i++) {
    minValues[i] = Float.MAX_VALUE;
    titles[i] = table.getString(0, i + 2);
    for(int j = 0; j < rows - 1; j++) {
      data[i][j] = table.getFloat(j + 1, i + 2);
     
      if(data[i][j] > maxValues[i]) maxValues[i] = data[i][j];
      else if(data[i][j] < minValues[i]) minValues[i] = data[i][j];
    }
  }
  
  for(int i = 0; i < startingMonth.length; i++) {
    for(int j = 0; j < rows - 1; j++) {
      if(data[i * 2][j] > 0 || data[i * 2 + 1][j] > 0) {
        
        startingMonth[i] = table.getInt(j + 1, 1);
        startingYear[i] = table.getInt(j + 1, 0);
        indicesForStartingMonth[i] = ((j - 1) >= 0)? j - 1 : 0;
        
        startDate[i] = table.getString(j + 1, 0) + "/" + table.getString(j + 1, 1);
        
        println(i  + " " +startDate[i] + " "+  titles[i * 2 + 1]);
        
        break;
      }
    }
  }
  
  spiralTimeLine = new ArrayList<PVector>();
  makeHalfCylinderMesh(30, 1, 1);
  addSpiral();
  
  
  
  Rotation rot =new Rotation(new Vector3D(0, 0, 1), 0);
  Vector3D center = new Vector3D(0, 0, 0);
  camStates[0] = new CameraState(rot, center, 2000);
  
  
  for(int i = 0; i < itemCount; i++) {
    PVector origin = new PVector(0, 0, startPosition[i].z);
    PVector direction = PVector.sub(startPosition[i], origin).normalize();;
    PVector newCenter = PVector.add(startPosition[i], direction.mult(boxWidth * 6));
    rot =new Rotation(new Vector3D(0, 0, 1), rotationAngle[i]);
    center = new Vector3D(newCenter.x, newCenter.y, newCenter.z);
    camStates[i + 1] = new CameraState(rot, center, 200);
  }
  
  
  cp5 = new ControlP5(this);
  
  selectGraphTypeRadioButton = cp5.addRadioButton("SelectGraph")
                              .setPosition(constantHorizontalMargin - 7, constantVerticalMargin / 2 + 45)
                              .setSize(20, 20)
                              .setColorForeground(color(120))
                              .setColorActive(color(#f9d423))
                              .setColorLabel(color(0))
                              .setItemsPerRow(2)
                              .setSpacingColumn(110);
                              
  selectGraphTypeRadioButton.addItem("Bar", 1);
  selectGraphTypeRadioButton.addItem("Half Disc", 2);
  
  for(Toggle t : selectGraphTypeRadioButton.getItems()) {
     t.getCaptionLabel().setColorBackground(color(255,80));
     t.getCaptionLabel().getStyle().moveMargin(-7,0,0,-3);
     t.getCaptionLabel().getStyle().movePadding(7,0,0,3);
     t.getCaptionLabel().getStyle().backgroundWidth = 90;
     t.getCaptionLabel().getStyle().backgroundHeight = 13;
     t.getCaptionLabel().setFont(createFont("Helvetica", 10));
  }
  
  selectGraphTypeRadioButton.activate(0);
  
  selectItemRadioButton = cp5.addRadioButton("SelectItem")
                              .setPosition(constantHorizontalMargin - 7, constantVerticalMargin / 2 + 80)
                              .setSize(20, 20)
                              .setColorForeground(color(120))
                              .setColorActive(color(#f9d423))
                              .setColorLabel(color(0))
                              .setItemsPerRow(1)
                              .setSpacingRow(1);

  selectItemRadioButton.addItem("Show All", 0);
  
  for(int i = 0; i < itemCount; i++) 
    selectItemRadioButton.addItem(titles[i * 2 + 1], i + 1);
  
  for(Toggle t : selectItemRadioButton.getItems()) {
     t.getCaptionLabel().setColorBackground(color(255,80));
     t.getCaptionLabel().getStyle().moveMargin(-7,0,0,-3);
     t.getCaptionLabel().getStyle().movePadding(7,0,0,3);
     t.getCaptionLabel().getStyle().backgroundWidth = 220;
     t.getCaptionLabel().getStyle().backgroundHeight = 13;
     t.getCaptionLabel().setFont(createFont("Helvetica", 10));
  }
  
  selectItemRadioButton.activate(0);
  
  cp5.setAutoDraw(false);
  
  textMode(SHAPE);
}


void controlEvent(ControlEvent theEvent) {
  int value = (int)theEvent.getValue();
  if(theEvent.isFrom(selectItemRadioButton)) {
    if(value >= 0) {
      selectedIndex = value;
      cam.setState(camStates[selectedIndex], 1000);
    }
  } else if(theEvent.isFrom(selectGraphTypeRadioButton)) {
    if(value >= 0) {
      switch(value) {
      case 1:
        mode = BAR_GRAPH;
        break;
      case 2:
        mode = CIRCLE_GRAPH;
        break;
      default:
        break;
      }
    }
  }
}

void gui() {
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  //translate();
  cp5.draw();
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
}

void addSpiral() {
  
  int years = (rows - 1) / 12;
  
  float heightPerYear = 50; 
  float daysPerMonth = 30;

  float zStep = heightPerYear / (daysPerMonth * 12); 
  float thetaStep = (2 * PI) / (daysPerMonth * 12); 
  
  float z = (years * heightPerYear) * 0.5;
  float r = 100;
  
  cam.setDistance(z);
    
  
  for(int i = 0; i < years; i++) {
    float theta = 0;
    
    
    for(int j = 0; j < daysPerMonth * 12; j++) {  
      float x = r * cos(theta);
      float y = r * sin(theta);
      
      spiralTimeLine.add(new PVector(x, y, z));
      
      z -= zStep;
      
      if((j % 30) == 0) {
        for(int k = 0; k < itemCount; k++) {  
          if(((j / 30 + 1) == startingMonth[k]) && i == (startingYear[k] - startYear)) {

            startPosition[k] = new PVector(x, y, z);
            rotationAngle[k] = theta;
            
          }
        }
      }
      
      theta += thetaStep;
    }
  }
}

void makeHalfCylinderMesh(int sides, float r, float h) {
  float theta = PI / (float)sides;
  float halfHeight = h / 2;
  
  for(int i = 0; i <= sides; i++) {
    float z = r * cos(i * theta);
    float y = r * sin(i * -theta);
    
    PVector top = new PVector(-halfHeight, y, z);
    PVector bottom = new PVector(halfHeight, y, z);
    halfCylinderMeshTop.add(top);
    halfCylinderMeshBottom.add(bottom);
    halfCylinderMeshBody.add(top);
    halfCylinderMeshBody.add(bottom);
  }
  
  halfCylinderMeshBody.add(halfCylinderMeshBody.get(0));
  halfCylinderMeshBody.add(halfCylinderMeshBody.get(1));
}

void drawHalfCylinder(float scaleX, float scaleY, float scaleZ, color c) {
  noStroke();
  
  pushMatrix();
    
    scale(scaleX, scaleY, scaleZ);
    
    fill(getNewHSBColor(c, 1.0, 1.0, 0.5));
    beginShape();
    
    for(PVector p: halfCylinderMeshTop) {
      vertex(p.x, p.y, p.z);
    }
    endShape(CLOSE);
    
    beginShape();
    
    for(PVector p: halfCylinderMeshBottom) {
      vertex(p.x, p.y, p.z);
    }
    endShape(CLOSE);
    
    
    fill(c);
    beginShape(TRIANGLE_STRIP);
    for(PVector p: halfCylinderMeshBody) {
      
      vertex(p.x, p.y, p.z);
    }
    endShape(CLOSE);
  popMatrix();
}

color getNewHSBColor(color c, float hRatio, float sRatio, float bRatio) {
  
  float h = hue(c) * hRatio;
  float s = saturation(c) * sRatio;
  float b = brightness(c) * bRatio;
  
  colorMode(HSB, 255);
  color newColor = color(h , s, b);
  colorMode(RGB, 255);
  return newColor;
}

void draw() {
  
  background(backgroundColor);
  
  stroke(200);
  noFill();
  
  
  pushMatrix();
  
  for(int i = 0; i < spiralTimeLine.size() - 1; i++) {
    PVector lineStart = spiralTimeLine.get(i);
    PVector lineEnd = spiralTimeLine.get(i + 1);
    
    line(lineStart.x, lineStart.y, lineStart.z, lineEnd.x, lineEnd.y, lineEnd.z);
  }
  
  for(int i = 0; i < startPosition.length; i++) {
    
    if(selectedIndex > 0 && selectedIndex - 1 != i) continue; 
      
      
    int monthIndex = indicesForStartingMonth[i];
    
    PVector lineStart = startPosition[i];
    
    pushMatrix();
      translate(lineStart.x, lineStart.y, lineStart.z);
      rotate(rotationAngle[i]);
      
      stroke(255);
      line(0, 0, 0, 50, 0, 0);
      
      textSize(10);
      
      pushMatrix();
        noStroke();
        float labelWidth = textWidth(titles[i * 2 + 1].toUpperCase());
        fill(0);
        translate(0, 0, -1);
        rect(-2,-boxMaxHeight -25,labelWidth+5, 12);
      popMatrix();
      
      fill(255);
      
      textAlign(LEFT, CENTER);
      text(titles[i * 2 + 1].toUpperCase(),0, -boxMaxHeight - 20);
      
      textSize(8);
      textAlign(LEFT, BOTTOM);
      text(startDate[i], 5, 0);
      
      noFill();
      
      translate(50, 0, 0);
      int year = startingYear[i];
      int month = startingMonth[i];
      
      for(; monthIndex < months - 1; monthIndex++) {
        if(month == 1 && year > startingYear[i]) {
          stroke(255);
          line(0, 0, 0, 0, boxMaxHeight + 20, 0);
          fill(255);
          textSize(10);
          textAlign(LEFT, TOP);
          text(year,3 , boxMaxHeight + 10);
        }
        
        noStroke();

        if(maxValues[i * 2] > 0) {
          float rectOSTHeight =  map(data[i * 2][monthIndex + 1], minValues[i * 2], maxValues[i * 2], 0, boxMaxHeight);
          color c = getNewHSBColor(itemColors[0], 1.0, 1.0, rectOSTHeight / boxMaxHeight);
          pushMatrix();
            if(mode == BAR_GRAPH) {
              //noFill();
              fill(c);
              translate(0, rectOSTHeight * -0.5, 0);
              box(boxWidth, rectOSTHeight, boxWidth);
            } else if(mode == CIRCLE_GRAPH) {
              drawHalfCylinder(boxWidth, rectOSTHeight, rectOSTHeight, c);//color(255 * rectOSTHeight / boxMaxHeight, 0, 0)
              
            }
            
            float mouseObjectDistance = sq(mouseX - screenX(0, 0, 0)) + sq(mouseY - screenY(0, 0, 0));
            
            if (mouseObjectDistance < 100) {
              
            }
          popMatrix();
        }
        
        if(maxValues[i * 2 + 1] > 0) {
          float rectMovieHeight = map(data[i * 2 + 1][monthIndex + 1], minValues[i * 2 + 1], maxValues[i * 2 + 1], 0, boxMaxHeight);
          color c = getNewHSBColor(itemColors[1], 1.0, 1.0, rectMovieHeight / boxMaxHeight);
          pushMatrix();
            if(mode == BAR_GRAPH) {
              //noFill();
              fill(c);
              translate(0, rectMovieHeight * 0.5, 0);
              box(boxWidth, rectMovieHeight, boxWidth);
            } else if(mode == CIRCLE_GRAPH) {
              drawHalfCylinder(boxWidth, -rectMovieHeight, rectMovieHeight, c);//color(255 * rectMovieHeight / boxMaxHeight));
            }
            float mouseObjectDistance = sq(mouseX - screenX(0, 0, 0)) + sq(mouseY - screenY(0, 0, 0));
            
            if (mouseObjectDistance < 100) {
              
            }
          popMatrix();
        }
        
        translate(boxWidth, 0, 0);
        
        month++;
        if(month == 13) {
          month = 1;
          year++;
        }
      }
    popMatrix();
  }
  
  popMatrix();
  
  fill(255);
  textSize(12);
  textAlign(CENTER);
  text("Timeline \nof Issue Date", 0, 0);
  
  
  cam.beginHUD();
  
  
  textSize(30);
  
  pushMatrix();
    noStroke();
    float labelWidth = textWidth(title);
    fill(color(0, 0, 0, 150));
    translate(0, 0, -1);
    rect(constantHorizontalMargin - 7, constantVerticalMargin / 2 - 20, labelWidth + 15, 45);
  popMatrix();
  
  fill(255);    
  textAlign(LEFT, CENTER);
  
  text(title, constantHorizontalMargin, constantVerticalMargin / 2);
  
  textSize(15);
  
  float yPosition = height - constantVerticalMargin;
  pushMatrix();
    noStroke();
    float legendWidth = 60 + 20 * boxWidth * 0.5 + textWidth(legendDescription);
    fill(color(0, 0, 0, 150));
    translate(0, 0, -1);
    rect(constantHorizontalMargin - 7, yPosition - boxMaxHeight * 0.5 - 7, legendWidth + 15, boxMaxHeight + 15);
  popMatrix();
  
  fill(255);
  
  textAlign(LEFT, TOP);
  
  text("DVD", constantHorizontalMargin, yPosition);
  textAlign(LEFT, BOTTOM);
  text("OST", constantHorizontalMargin, yPosition);
  
  noStroke();
  for(int i = 0; i < 20; i++) {
    float barHeight= boxMaxHeight * 0.5 * i / 20 + 0.1;
    color c = getNewHSBColor(itemColors[1], 1.0, 1.0, i / 20.0);
    fill(c);
    rect(constantHorizontalMargin + 50 + i * boxWidth * 0.5, yPosition, boxWidth * 0.5, barHeight);
    
    c = getNewHSBColor(itemColors[0], 1.0, 1.0, i / 20.0);
    fill(c);
    rect(constantHorizontalMargin + 50 + i * boxWidth * 0.5, yPosition - barHeight, boxWidth * 0.5, barHeight);
  }
  
  fill(255);
  textSize(15);
  textAlign(LEFT, CENTER);
  text(legendDescription, constantHorizontalMargin + 60 + 20 * boxWidth * 0.5, yPosition);
  cam.endHUD();
    
  gui();
}

int camIndex = 0; 
void keyPressed() {
  
  switch(key) {
    case '0':
      shapeTextMode = !shapeTextMode;
      
      if(shapeTextMode) textMode(SHAPE);
      else textMode(MODEL);
  
      break;
    case '1':
      mode = BAR_GRAPH;
      break;
    case '2':
      mode = CIRCLE_GRAPH;
      break;
    case '3':
      cam.setState(camStates[camIndex], 1000);
      camIndex = (camIndex + 1) % camStates.length;
      break;
    
    default:
      break;
  }

}